@@ -36,6 +36,14 @@ void process_file(NODE *match_pattern,char *fn,int fn_flag) {
    char *input_buffer=NULL;
    int inbuf_size=0,inbuf_used=0,parse_ptr=0;
    FILE *infile;
+   HASHED_STRING *hfn,*colon;
+   
+   /* wrap the filename in a string so we can escape-print it */
+   if (fn_flag>=0)
+     hfn=new_string(strlen(fn+fn_flag),fn+fn_flag);
+   else
+     hfn=new_string(strlen(fn),fn);
+   colon=new_string(1,":");
    
    /* open input file */
    if (strcmp(fn,"-")) {
@@ -89,14 +97,18 @@ void process_file(NODE *match_pattern,char *fn,int fn_flag) {
 	    if (tree_match(match_pattern,to_match)) {
 	       for (i=0;((unsigned char)input_buffer[i])<=0x20;i++);
 	       if (fn_flag>=0)
-		 printf(":%s:",fn+fn_flag);
-	       fwrite(input_buffer+i,1,parse_ptr-i,stdout);
-	       echoing_whitespace=1;
-	       
+		 write_bracketed_string(hfn,colon);
+	       if (cook_output)
+		 write_cooked_tree(to_match);
+	       else {
+		  fwrite(input_buffer+i,1,parse_ptr-i,stdout);
+		  echoing_whitespace=1;
+	       }
 	    }
 	    free_node(to_match);
 	    if (parse_ptr<inbuf_used)
-	      memmove(input_buffer,input_buffer+parse_ptr,inbuf_used-parse_ptr);
+	      memmove(input_buffer,input_buffer+parse_ptr,
+		      inbuf_used-parse_ptr);
 	    inbuf_used-=parse_ptr;
 	    parse_ptr=0;
 	 } else
@@ -108,11 +120,14 @@ void process_file(NODE *match_pattern,char *fn,int fn_flag) {
      fclose(infile);
    
    free(input_buffer);
+   delete_string(hfn);
+   delete_string(colon);
 }
 
 /**********************************************************************/
 
 static struct option long_opts[] = {
+   {"cooking",required_argument,NULL,'c'},
    {"dictionary",optional_argument,NULL,'d'},
    {"help",no_argument,NULL,'h'},
    {"version",no_argument,NULL,'V'},
@@ -141,13 +156,17 @@ int main(int argc,char **argv) {
    register_syntax();
 
    /* loop on command-line options */
-   while ((c=getopt_long(argc,argv,"Vd::h",long_opts,NULL))!=-1) {
+   while ((c=getopt_long(argc,argv,"Vc:d::h",long_opts,NULL))!=-1) {
       switch (c) {
 
        case 'V':
 	 show_version=1;
 	 break;
 
+       case 'c':
+	 set_output_recipe(optarg);
+	 break;
+	 
        case 'd':
 	 if (optarg==NULL)
 	   dictname="";
@@ -177,6 +196,7 @@ int main(int argc,char **argv) {
 	  "PATTERN should be an Extended Ideographic Description Sequence\n\n"
 	  "Options:\n"
 	  "  -V, --version             display version and license\n"
+	  "  -c, --cooking=FMT         set input/output cooking\n"
 	  "  -d, --dictionary=NAME     search standard dictionary\n"
 	  "  -h, --help                display this help");
    
@@ -208,7 +228,7 @@ int main(int argc,char **argv) {
       sprintf(dictglob,"%s/%s*.eids",dictdir,dictname);
       if (glob(dictglob,0,NULL,&globres)==0) {
 	 num_files+=globres.gl_pathc;
-	 for(c=0;c<globres.gl_pathc;c++)
+	 for (c=0;c<globres.gl_pathc;c++)
 	   process_file(match_pattern,globres.gl_pathv[c],
 			num_files>1?strlen(dictdir)+1:-1);
 	 globfree(&globres);
@@ -219,9 +239,13 @@ int main(int argc,char **argv) {
    while (optind<argc)
      process_file(match_pattern,argv[optind++],num_files>1?0:-1);
 
-   /* read stdin if no other choices */
-   if ((num_files==0) && (dictname==NULL))
-     process_file(match_pattern,"-",-1);
+   /* read stdin or complain */
+   if (num_files==0) {
+      if (dictname==NULL)
+	process_file(match_pattern,"-",-1);
+      else
+	puts("(no dictionaries were searched)");
+   }
    
    exit(0);
 }
