@@ -21,6 +21,10 @@
 
 #include "config.h"
 
+#ifdef HAVE_PCRE
+#include <pcre.h>
+#endif
+
 /**********************************************************************/
 
 typedef struct _NODE *(*MATCH_FN)(struct _NODE *);
@@ -40,11 +44,26 @@ typedef struct _HASHED_STRING {
    size_t length;
    int refs,arity;
    MATCH_FN match_fn;
+#ifdef HAVE_PCRE
+   pcre *pcre_compiled;
+   pcre_extra *pcre_studied;
+#endif
 } HASHED_STRING;
 
 typedef struct _NODE {
    HASHED_STRING *head,*functor;
-   struct _NODE *child[3],*match_parent;
+#ifdef ANON_UNION_STRUCT
+   union {
+      struct _NODE *child[3];
+      struct { struct _NODE *nc_next,*nc_needle,*nc_haystack; };
+   };
+#else
+   struct _NODE *child[3];
+#define nc_next child[0]
+#define nc_needle child[1]
+#define nc_haystack child[2]
+#endif
+   struct _NODE *match_parent;
    int refs,arity,complete;
    MATCH_RESULT match_result;
 } NODE;
@@ -63,6 +82,22 @@ typedef enum _PARSE_STATE {
 
 /**********************************************************************/
 
+/* assoc.c */
+
+NODE *assoc_match_fn(NODE *);
+
+/**********************************************************************/
+
+/* cook.c */
+
+extern int cook_output,canonicalize_input;
+
+void set_output_recipe(char *);
+void write_bracketed_string(HASHED_STRING *,HASHED_STRING *);
+void write_cooked_tree(NODE *);
+
+/**********************************************************************/
+
 /* hash.c */
 
 HASHED_STRING *new_string(size_t,char *);
@@ -79,10 +114,8 @@ NODE *default_match_fn(NODE *);
 NODE *and_or_match_fn(NODE *);
 NODE *anything_match_fn(NODE *);
 NODE *anywhere_match_fn(NODE *);
-NODE *assoc_match_fn(NODE *);
 NODE *equal_match_fn(NODE *);
 NODE *not_match_fn(NODE *);
-NODE *regex_match_fn(NODE *);
 NODE *unord_match_fn(NODE *);
 
 int tree_match(NODE *,NODE *);
@@ -91,6 +124,8 @@ int tree_match(NODE *,NODE *);
 
 /* parse.c */
 
+extern HASHED_STRING *hashed_bracket[15];
+
 extern NODE **parse_stack;
 extern int stack_ptr;
 extern PARSE_STATE parse_state;
@@ -98,3 +133,9 @@ extern int echoing_whitespace;
 
 size_t parse(size_t,char *);
 void register_syntax(void);
+
+/**********************************************************************/
+
+/* regex.c */
+
+NODE *regex_match_fn(NODE *);
